(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010 Steffen Solyga <solyga@absinth.net>                               *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Put, Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Comp IMPORT TValue, TOperation, Val, Opr, Sto, Lod;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;
  vFalse = 0;
  vTrue  = 1;

TYPE
  TClass = ( cConst, cVar, cType );
  TType = ( tBool, tInt );
  TObjDesc = RECORD
    id : CARDINAL;
    typ: TType;
    CASE cls: TClass OF
    | cConst: val: TValue;
    | cVar  : idx: CARDINAL;
    | cType :
    END;
  END;

  TMode = ( mConst, mVar, mStack );
  TItemDesc = RECORD
    t: TType;
    CASE c: BOOLEAN OF (* const *)
    | TRUE : v: TValue;
    | FALSE: (* stack *)
    END;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;
  idx: CARDINAL;
  oprs: ARRAY TSymbol OF TOperation;
  xops: ARRAY TSymbol OF TOperation;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Find( VAR i: CARDINAL );
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].id, id ) = 0 THEN RETURN END;
    INC( i );
  END;
END Find;

PROCEDURE True( v: TValue ): BOOLEAN;
BEGIN
  RETURN v = vTrue;
END True;

PROCEDURE Value( x: BOOLEAN ): TValue;
BEGIN
  IF x THEN RETURN vTrue ELSE RETURN vFalse END;
END Value;


PROCEDURE Expression( VAR d: TItemDesc );

  PROCEDURE SimpleExpr( VAR d: TItemDesc );

    PROCEDURE Factor( VAR d: TItemDesc );
    VAR i: CARDINAL;
    BEGIN
      CASE sym OF
      | sNum : IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
               d.t := tInt; d.c := TRUE; d.v := num;
               Read();
      | sId  : Find( i );
               IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
               d.t := objs[ i ].typ;
               CASE objs[ i ].cls OF
               | cConst: d.c := TRUE; d.v := objs[ i ].val;
               | cVar  : d.c := FALSE; Lod( objs[i].idx );
               ELSE
                 Err( 'Expected constant or variable' );
               END;
               Read();
      | sLPar: Read(); Expression( d );
               IF sym # sRPar THEN Err( 'Expected operator or ")"' ) END;
               Read();
      | sNot : Read(); Factor( d );
               IF d.t # tBool THEN Err( 'Expected boolean factor' ) END;
               IF d.c THEN
                 d.v := Value( ~True(d.v) );
               ELSE
                 Opr( oprs[sNot] );
               END;
      ELSE
        Err( 'Expected factor' );
      END;
    END Factor;

    PROCEDURE Term( VAR d: TItemDesc );
    VAR s: TSymbol; d1: TItemDesc;
    BEGIN
      Factor( d );
      WHILE ( sMul <= sym ) & ( sym <= sAnd ) DO
        IF ( sym = sAnd ) # ( d.t = tBool ) THEN
          Err( 'Wrong operator type' )
        END;
        s := sym; Read(); Factor( d1 );
        IF d.t # d1.t THEN Err( 'Wrong operand type' ) END;
        IF d.c & d1.c THEN
          CASE s OF
          | sMul: d.v := d.v  *  d1.v;
          | sDiv: d.v := d.v DIV d1.v;
          | sAnd: d.v := Value( True(d.v) & True(d1.v) );
          ELSE
            Err( 'Bug: Invalid mulop' );
          END;
        ELSE
          IF d.c THEN
            Val( d.v ); Opr( xops[s] );
          ELSE
            IF d1.c THEN Val( d1.v ) END;
            Opr( oprs[s] );
          END;
          d.c := FALSE;
        END;
      END;
    END Term;

  VAR s: TSymbol; d1: TItemDesc;
  BEGIN
    IF ( sym = sPlus ) OR ( sym = sMinus ) THEN
      s := sym; Read(); Term( d );
      IF d.t # tInt THEN Err( 'Expected integer term' ) END;
      IF s = sMinus THEN
        IF d.c THEN d.v := -d.v ELSE Opr( oNeg ) END;
      END;
    ELSE
      Term( d );
    END;
    WHILE ( sPlus <= sym ) & ( sym <= sOr ) DO
      IF ( sym = sOr ) # ( d.t = tBool ) THEN Err( 'Wrong operator type' ) END;
      s := sym; Read(); Term( d1 );
      IF d.t # d1.t THEN Err( 'Wrong operand type' ) END;
      IF d.c & d1.c THEN
        CASE s OF
        | sPlus : d.v := d.v + d1.v;
        | sMinus: d.v := d.v - d1.v;
        | sOr   : d.v := Value( True(d.v) OR True(d1.v) );
        ELSE
          Err( 'Bug: Invalid addop' );
        END;
      ELSE
        IF d.c THEN
          Val( d.v ); Opr( xops[s] );
        ELSE
          IF d1.c THEN Val( d1.v ) END;
          Opr( oprs[s] );
        END;
        d.c := FALSE;
      END;
    END;
  END SimpleExpr;

VAR s: TSymbol; d1: TItemDesc;
BEGIN
  SimpleExpr( d );
  IF ( sEql <= sym ) & ( sym <= sGeq ) THEN
    s := sym; Read(); SimpleExpr( d1 );
    IF d.t # d1.t THEN Err( 'Wrong operand type' ) END;
    d.t := tBool;
    IF d.c & d1.c THEN
      CASE s OF
      | sEql: d.v := Value( d.v =  d1.v );
      | sNeq: d.v := Value( d.v #  d1.v );
      | sLss: d.v := Value( d.v <  d1.v );
      | sLeq: d.v := Value( d.v <= d1.v );
      | sGtr: d.v := Value( d.v >  d1.v );
      | sGeq: d.v := Value( d.v >= d1.v );
      ELSE
        Err( 'Bug: Invalid relop' );
      END;
    ELSE
      IF d.c THEN
        Val( d.v ); Opr( xops[s] );
      ELSE
        IF d1.c THEN Val( d1.v ) END;
        Opr( oprs[s] );
      END;
      d.c := FALSE;
    END;
  END;
END Expression;

PROCEDURE Statement();
VAR i: CARDINAL; d: TItemDesc;
BEGIN
  CASE sym OF
  | sId : (* assignment *)
          Find( i );
          IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
          IF objs[ i ].cls # cVar THEN Err( 'Not a variable' ) END;
          Read();
          IF sym # sBec THEN Err( 'Expected ":="' ) END;
          Read(); Expression( d );
          IF d.t # objs[i].typ THEN Err( 'Incompatible types' ) END;
          IF d.c THEN Val( d.v ) END;
          Sto( objs[i].idx );
  | sWrt: (* expression output *)
          Read(); Expression( d );
          IF d.c THEN Val( d.v ) END;
          CASE d.t OF
          | tInt : Opr( oWrt );
          | tBool: Opr( oWrB );
          ELSE
            Err( 'Bug: Invalid type' );
          END;
  ELSE
    (* empty *)
  END;
END Statement;

PROCEDURE ConstDeclaration();
VAR i: CARDINAL; d: TItemDesc;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cConst;
  KeepId();
  Read();
  IF sym # sEql THEN Err( 'Expected "="' ) END;
  Read(); Expression( d );
  IF ~ d.c THEN Err( 'Expected constant' ) END;
  objs[ i ].typ := d.t;
  objs[ i ].val := d.v;
  INC( cnt );
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END ConstDeclaration;

PROCEDURE VarDeclaration();
VAR i,j: CARDINAL;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cVar;
  KeepId();
  Read();
  IF sym # sCol THEN Err( 'Expected ":"' ) END;
  Read();
  IF sym # sId THEN Err( 'Expected identifier' ) END;
  Find( j );
  IF j >= cnt THEN Err( 'Identifier undeclared' ) END;
  IF objs[ j ].cls # cType THEN Err( 'Not a type' ) END;
  objs[ i ].typ := objs[ j ].typ;
  objs[ i ].idx := idx;
  INC( cnt ); INC( idx );
  Read();
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END VarDeclaration;

PROCEDURE Run(); (* Program() *)
BEGIN
  Read();
  IF sym = sConst THEN
    Read();
    WHILE sym = sId DO ConstDeclaration() END;
  END;
  IF sym = sVar THEN
    Read();
    WHILE sym = sId DO VarDeclaration() END;
  END;
  IF sym # sBeg THEN Err( 'Expected "BEGIN"' ) END;
  Read(); Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
  IF sym # sDot THEN Err( 'Expected ";" or "."' ) END;
(*  Read(); IF sym # sEOF THEN Err( 'Expected EOF' ) END; *)
END Run;

PROCEDURE InitOprs();
VAR s: TSymbol;
BEGIN
  FOR s := MIN( TSymbol ) TO MAX( TSymbol ) DO oprs[ s ] := oNull END;
  oprs[ sWrt   ] := oWrt;
  oprs[ sNot   ] := oNot;
  oprs[ sMul   ] := oMul;
  oprs[ sDiv   ] := oDiv;
  oprs[ sAnd   ] := oAnd;
  oprs[ sPlus  ] := oAdd;
  oprs[ sMinus ] := oSub;
  oprs[ sOr    ] := oOr;
  oprs[ sEql   ] := oEql;
  oprs[ sNeq   ] := oNeq;
  oprs[ sLss   ] := oLss;
  oprs[ sLeq   ] := oLeq;
  oprs[ sGtr   ] := oGtr;
  oprs[ sGeq   ] := oGeq;
  FOR s := MIN( TSymbol ) TO MAX( TSymbol ) DO xops[ s ] := oNull END;
  xops[ sWrt   ] := oWrt;
  xops[ sNot   ] := oNot;
  xops[ sMul   ] := oMul;
  xops[ sDiv   ] := oVid;
  xops[ sAnd   ] := oAnd;
  xops[ sPlus  ] := oAdd;
  xops[ sMinus ] := oBus;
  xops[ sOr    ] := oOr;
  xops[ sEql   ] := oEql;
  xops[ sNeq   ] := oNeq;
  xops[ sLss   ] := oGtr;
  xops[ sLeq   ] := oGeq;
  xops[ sGtr   ] := oLss;
  xops[ sGeq   ] := oLeq;
END InitOprs;

BEGIN
  cnt := 0;
  idx := 0;
  Put( 'FALSE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vFalse;
  KeepId(); INC( cnt );
  Put( 'TRUE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := vTrue;
  KeepId(); INC( cnt );
  Put( 'BOOLEAN' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  Put( 'INTEGER' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tInt;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  InitOprs();
END Prsr.
